#!/usr/bin/php
<?php
// Author: Philip Eklöf <peklof@op5.com> 2014-09-10
// https://bugs.op5.com/view.php?id=9016 workaround

class Base
{
	protected $pdo = NULL;
	public $debug = false;
	public $db_ts = array();
	public $db_state = array();
	public $log_ts = array();
	public $host_inserts = 0;
	public $service_inserts = 0;

	public function __construct()
	{
		set_error_handler(array($this, 'exception_error_handler'));
	}

	public function exception_error_handler($errno, $errstr, $errfile, $errline)
	{
		throw new ErrorException($errstr, $errno, 0, $errfile, $errline);
	}

	public function exception_parse($e)
	{
		$ret = array();
		do
		{
			$trace = $e->getTrace();
			$func = $trace[0]['function'];
			$ret[] = array(
				'file' => $e->getFile(),
				'line' => $e->getLine(),
				'func' => $func,
				'msg' => $e->getMessage()
			);
		}
		while($e = $e->getPrevious());

		return $ret;
	}

	public function exception_die($e)
	{
		$trace = $this->exception_parse($e);
		foreach(array_reverse($trace) as $n => $a)
		{
			printf(
				"#%d %s@%s %s(): %s\n",
				$n, $a['file'], $a['line'], $a['func'], $a['msg']
			);
		}
		exit();
	}

	public function debug($msg, $debug=false)
	{
		if($this->debug)
			$debug = true;
		if($debug)
			echo "$msg\n\n";
		return NULL;
	}

	public function msg($msg)
	{
		echo date('[H:m:s] ')."$msg\n";
	}

	public function db_init($db_settings=array())
	{
		if(!is_array($db_settings))
			throw new Exception('PARAM_NOT_ARRAY');

		$names = array('host', 'name', 'pass', 'user', 'type');
		foreach($names as $name)
			if(!isset($db_settings[$name]))
				throw new Exception('MISSING_SETTING:' . $name);

		$t = $db_settings['type'];
		$h = $db_settings['host'];
		$n = $db_settings['name'];

		try {
			$this->pdo = new PDO(
				"{$t}:host={$h};dbname=${n}",
				$db_settings['user'],
				$db_settings['pass']
			);
		} catch(PDOException $e) {
			throw new Exception('PDO_CONNECT_ERROR', 0, $e);
		}

		if(!$this->pdo->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_EXCEPTION))
			throw new Exception('PDO_SET_ATTRIBUTE_ERROR');

		return true;
	}

	protected function db_sth_prep_exec($query, $params=array())
	{
		if(!is_object($this->pdo))
			throw new Exception('PDO_NOT_INITIALIZED');

		// prepare query
		try {
			$sth = $this->pdo->prepare($query);
		} catch(PDOException $e) {
			throw new Exception('PDO_PREPARE', 0, $e);
		}

		// execute query
		try {
			$sth->execute($params);
		} catch(PDOException $e) {
			throw new Exception('PDO_EXECUTE', 0, $e);
		}

		return $sth;
	}

	public function get_db_initials()
	{
		// by using min(timestamp) and GROUP BY, only the oldest entry of each
		// group is selected.
		$query = '
			SELECT
				host_name,
				service_description,
				state,
				min(timestamp) AS ts
			FROM
				report_data
			WHERE
				event_type = 701
				OR
				event_type = 801
			GROUP BY
				host_name,
				service_description
		';

		// prepare and exec query
		try {
			$sth = $this->db_sth_prep_exec($query);
		} catch(PDOException $e) {
			throw new Exception('ERROR_INSERTING_ENTRY', $e);
		}

		$this->db_ts = array();;
		$this->db_state = array();
		foreach($sth->fetchAll(PDO::FETCH_ASSOC) as $r)
		{
			if(empty($r['host_name']))
				continue;

			$host = base64_encode($r['host_name']);
			$service = base64_encode($r['service_description']);
			$this->db_ts[$host][$service] = $r['ts'];
			$this->db_state[$host][$service] = $r['state'];
		}
	}

	public function get_log_initials()
	{
		$cmd = 'bash -c "';
		$cmd .= 'export LANG=C; mon log merge naemon | ';
		$cmd .= "egrep -o '^\[[0-9]+\] INITIAL (HOST STATE: [^;]+;|SERVICE STATE: [^;]+;[^;]+;)' | ";
		$cmd .= 'sort -k5 | uniq -f2';
		$cmd .= '"';
		// 1. LANG=C to remain sane.
		// 2. Let "mon log merge" find and output all nagios/naemon logs for us.
		// 3. Select all host and service initial state entries.
		// 4. Sort the data starting at key 5, i.e. group by host and host;service.
		//    The sorting automatically sorts each group by timestamp as well,
		//    oldest first (given they're more recent than 2001-09-09 01:46:39 :)).
		// 5. Finally, the uniq cmd ignores the timestamp of each line, resulting
		//    in all entries but the first of every group being stripped.

		exec($cmd, $res, $ret);
		if($ret !== 0)
			throw new Exception("FAILED_CMD:$ret");

		$regex = '/^\[([0-9]+)\] INITIAL (HOST STATE: ([^;]+);|SERVICE STATE: ([^;]+);([^;]+);)/';
		$this->log_ts = array();
		foreach($res as $r)
		{
			if(!preg_match($regex, $r, $matches))
				continue;

			if(isset($matches[4])) // service
			{
				// avoid black magic by base64 encoding the names
				$host = base64_encode($matches[4]);
				$service = base64_encode($matches[5]);
			}
			else // host
			{
				$host = base64_encode($matches[3]);
				$service = '';
			}

			$this->log_ts[$host][$service] = $matches[1];
		}
	}

	public function get_ls_hosts()
	{
		$hosts = array();
		exec("mon query ls hosts -c name", $hosts, $ret);
		if($ret !== 0)
			throw new Exception("FAILED_CMD:$ret");
		return $hosts;
	}

	public function get_ls_services($host)
	{
		if(!isset($host))
			throw new Exception('MISSING_PARAM');

		$host = escapeshellarg($host);

		$services = array();
		exec("mon query ls services -c description host_name=$host", $services, $ret);
		if($ret !== 0)
			throw new Exception("FAILED_CMD:$ret");
		return $services;
	}

	public function insert($ts, $host, $service='')
	{
		if(!isset($ts))
			throw new Exception('MISSING_PARAM:ts');
		if(!isset($host))
			throw new Exception('MISSING_PARAM:host');

		$date = date('Y-m-d H:i:s', $ts);
		$this->debug("Inserting OK state into DB @ $date ($ts).", true);

		if($service === '') // host
		{
			$event_type = 801;
			$this->host_inserts++;
		}
		else // service
		{
			$event_type = 701;
			$this->service_inserts++;
		}

		$query = '
			INSERT INTO report_data
			(
				host_name,
				service_description,
				timestamp,
				state,
				hard,
				downtime_depth,
				event_type
			)
			VALUES
			(
				:entry_host,
				:entry_service,
				:entry_ts,
				0,
				1,
				0,
				:entry_event_type
			)
		';

		$params = array(
			'entry_host' => $host,
			'entry_service' => $service,
			'entry_ts' => $ts,
			'entry_event_type' => $event_type
		);

		// prepare and exec query
		try {
			$sth = $this->db_sth_prep_exec($query, $params);
		} catch(PDOException $e) {
			throw new Exception('ERROR_INSERTING_ENTRY', $e);
		}
	}

	public function verify($host, $service='')
	{
		if(!isset($host))
			throw new Exception('MISSING_PARAM');

		$host = base64_encode($host);
		$service = base64_encode($service);

		if(!isset($this->log_ts[$host][$service]))
			return $this->debug('Ignoring - not found in nagios logs.');

		// object not found in report_data at all, return timestamp to insert at once.
		if(!isset($this->db_ts[$host][$service]))
			return $this->log_ts[$host][$service];

		if($this->db_state[$host][$service] === '0')
			return $this->debug('Ignoring - first entry in DB is already OK.');

		$db_ts = $this->db_ts[$host][$service];
		$log_ts = $this->log_ts[$host][$service];

		if($log_ts >= ($db_ts - 300))
			return $this->debug("Ignoring - nagios.log entry ($log_ts) >= DB entry ($db_ts - 300).");

		// nagios.log entry is "much" older than the report_data entry, insert it!
		return $log_ts;
	}

}

function syntax($code=0)
{
	echo "Usage: op5 mt report-data-fixmisinit [--debug]\n";
	exit($code);
}


$o = new Base;

if(!is_object($o))
	die("MAIN_OBJECT_NOT_CREATED\n");

if(isset($argv[1]))
{
	if($argv[1] === '-h' || $argv[1] === '--help')
		syntax(0);

	if($argv[1] === '--debug')
		$o->debug = true;
	else
		syntax(1);
}

$db_settings = array(
	'host' => 'localhost',
	'user' => 'root',
	'pass' => '',
	'name' => 'merlin',
	'type' => 'mysql',
	'timeout' => 10
);

try {
	$o->db_init($db_settings);

	$o->msg('Fetching initial state information from report_data db table...');
	$o->get_db_initials();

	$o->msg('Fetching initial state information from nagios logs...');
	$o->get_log_initials();

	echo "\n";

	foreach($o->get_ls_hosts() as $host)
	{
		$o->msg("($host)");
		if(($ts = $o->verify($host)) !== NULL)
			$o->insert($ts, $host);

		foreach($o->get_ls_services($host) as $service)
		{
			$o->msg("($host;$service)");
			if(($ts = $o->verify($host, $service)) !== NULL)
				$o->insert($ts, $host, $service);
		}
	}

	echo "\n";
	$msg = 'All done. ';
	$msg .= $o->host_inserts . ' host(s) and ';
	$msg .= $o->service_inserts . ' service(s) ';
	$msg .= 'were inserted into the database.';
	$o->msg($msg);

} catch(Exception $e) {
	$o->exception_die($e);
}


?>
